@@ -32,6 +32,44 @@
  *          ______________________________________.
  *
  * $Log: main.cxx,v $
+ * Revision 1.46  2003/02/26 02:49:07  robertj
+ * Fixed race condition when makeing calls.
+ * Fixed caller hanging up issue, thanks Pietro Ravasio
+ *
+ * Revision 1.45  2003/01/06 06:23:39  robertj
+ * Increased maximum possible jitter configuration to 10 seconds.
+ * Fixed ability to set jitter size (condition wrong way round).
+ * Added abort ability and removed 100% CPU problem diring DTMF digit
+ *   gathering and awaiting connect.
+ *
+ * Revision 1.44  2002/12/11 10:48:36  rogerh
+ * delete listener when exiting
+ * Add NAT support with the --translate option (copied from ohphone)
+ *
+ * Revision 1.43  2002/11/10 08:12:42  robertj
+ * Moved constants for "well known" ports to better place (OPAL change).
+ *
+ * Revision 1.42  2002/10/31 00:55:13  robertj
+ * Enhanced jitter buffer system so operates dynamically between minimum and
+ *   maximum values. Altered API to assure app writers note the change!
+ * Removed some redundent code.
+ *
+ * Revision 1.41  2002/07/02 03:24:09  dereks
+ * Add LOTS of PTrace statements..  Improve check on LID ringing.
+ *
+ * Revision 1.40  2002/06/17 23:26:13  dereks
+ * Add fast start disable (-F) and H245 Tunnelling disable (-T) options.
+ *
+ * Revision 1.39  2002/03/20 06:09:39  robertj
+ * Added options for setting volume, thanks Jason Spence
+ * Bug fix for collecting digits time out, thanks David Rowe
+ *
+ * Revision 1.38  2002/01/27 14:37:10  rogerh
+ * Add --listenport option
+ *
+ * Revision 1.37  2001/11/19 06:42:34  robertj
+ * Added ability to collect digits before making H.323 connection, thanks David Rowe.
+ *
  * Revision 1.36  2001/06/20 07:36:57  robertj
  * Added dial plan parameter, thanks Chih-Wei Huang
  *
@@ -247,12 +285,15 @@ void PSTNGw::Main()
 #endif
              "c-calling-tone."
              "C-country:"
+	     "-collect."
              "d-dial:"
              "D-disable:"
              "h-help."
              "i-interface:"
              "f-force-dial."
+             "F-fast-disable."
              "j-jitter:"
+             "-listenport:"
              "n-numframes:"
 #if PTRACING
              "o-output:"
@@ -265,21 +306,27 @@ void PSTNGw::Main()
 #if PTRACING
              "t-trace."
 #endif
+             "T-h245tunneldisable."
              "-tos:"
+             "-translate:"
              "u-user:"
              "g-gatekeeper:"   
              "G-gatekeeper-id:"
              "-no-gatekeeper."
              "r-require-gatekeeper."
+	     "S-disable-h245-in-setup."
              "s-prefix:"
 #ifdef HAS_VPB
              "v-vpb:"
 #endif
+	     "-recordgain:"
+	     "-playgain:"
           , FALSE);
 
 #if PTRACING
   PTrace::Initialise(args.GetOptionCount('t'),
-                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL);
+                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL,
+		     PTrace::Blocks | PTrace::Timestamp | PTrace::Thread | PTrace::FileAndLine);
 #endif
 
   if (args.HasOption('h')) {
@@ -289,20 +336,28 @@ void PSTNGw::Main()
             "  -q --quicknet dev  : Use Quicknet LineJACK\n"
 #endif
 #ifdef HAS_VPB
-            "  -v --vpb dev       : Use Voicetronics VPB4\n"
+            "  -v --vpb dev       : Use Voicetronix VPB4\n"
 #endif
+            "     --playgain vol  : Set VPB4 playback gain\n"
+            "     --recordgain vol: Set VPB4 record gain\n"
             "  -u --user str      : Set local endpoint name\n"
             "  -n --numframes n   : Number of audio frames per ethernet packet\n"
 
             "  -i --interface ip  : Use specific interface\n"
+            "  --listenport       : Port to listen on for incoming connections (default 1720)\n"
             "     --tos n         : Set IP Type of Service byte to n\n"
             "  -j --jitter delay  : Set jitter buffer to delay milliseconds\n"
             "  -d --dial phone    : Default number to call\n"
+            "     --collect       : Collect digits before making call\n"
+            "  -F --fast-disable  : Disable fast start\n"
             "  -f --force-dial    : Force dial number in call\n"
             "  -c --calling-tone  : Enable PSTN calling tones translation\n"
             "  -C --country code  : Set the country settings for PSTN card\n"
             "  -D --disable codec : Disable the specified codec (may be used multiple times)\n"
-            "  -p --prefer codec  : Prefer the specified codec (may be used multiple times)\n"
+            "  -P --prefer codec  : Prefer the specified codec (may be used multiple times)\n"
+	    "  -S --disable-h245-in-setup : Disable H245 in setup\n"
+            "  -T --h245tunneldisable     : Disable H245 tunnelling\n"
+            "  --translate up             : Set external IP address to up is masQueraded\n"
 #ifdef HAS_IXJ
             "  -A --aec num       : Set AEC (default to medium)\n"
 #endif
@@ -311,11 +366,11 @@ void PSTNGw::Main()
             "  -o --output        : File for trace output, default is stderr\n"
 #endif
             "\nGatekeeper options:\n"
-            "  -g --gatekeeper host    : Specify gatekeeper host\n"
-            "  -G --gatekeeper-id name : Specify gatekeeper by ID\n"
-            "     --no-gatekeeper      : Disable gatekeeper discovery\n"
-            "  -r --require-gatekeeper : Exit if gatekeeper discovery fails\n"
-            "  -s --prefix str         : Set gateway supported prefix\n"
+            "  -g --gatekeeper host       : Specify gatekeeper host\n"
+            "  -G --gatekeeper-id name    : Specify gatekeeper by ID\n"
+            "     --no-gatekeeper         : Disable gatekeeper discovery\n"
+            "  -r --require-gatekeeper    : Exit if gatekeeper discovery fails\n"
+            "  -s --prefix str            : Set gateway supported prefix\n"
             "\n"
             "     --save          : Save arguments in configuration file\n"
             "  -h --help          : Display this help message\n";
@@ -350,21 +405,35 @@ BOOL GatewayEndPoint::Initialise(PArgList & args)
 {
   PINDEX i;
 
-  autoDial = args.GetOptionString('d');
-  forceDial = args.HasOption('f');
-  convertTones = args.HasOption('c');
+  autoDial         = args.GetOptionString('d');
+  forceDial        = args.HasOption('f');
+  convertTones     = args.HasOption('c');
+  disableFastStart = args.HasOption('f');
+  disableH245Tunneling = args.HasOption('T');
+  disableH245inSetup = args.HasOption('S');
 
-  jitter = 150;   /*Default jitter buffer length is 150ms   */
   if (args.HasOption('j')) {
-    unsigned newjitter = args.GetOptionString('j').AsUnsigned();
-    if (newjitter >= 20 && newjitter <= 10000)
-      jitter = newjitter;
+    unsigned minJitter;
+    unsigned maxJitter;
+    PStringArray delays = args.GetOptionString('j').Tokenise(",-");
+    if (delays.GetSize() == 2) {
+      minJitter = delays[0].AsUnsigned();
+      maxJitter = delays[1].AsUnsigned();
+    }
     else {
-      cerr << "Jitter should be between 20 milliseconds and 10 seconds." << endl;
-      cerr << "Using default value of 150 ms." << endl;
+      maxJitter = delays[0].AsUnsigned();
+      minJitter = PMIN(GetMinAudioJitterDelay(), maxJitter);
+    }
+    if (minJitter >= 20 && minJitter <= maxJitter && maxJitter <= 10000)
+      SetAudioJitterDelay(minJitter, maxJitter);
+    else {
+      cerr << "Jitter should be between 20 milliseconds and 10 seconds.\n"
+              "Using default value of " << GetMinAudioJitterDelay() <<
+              " to " << GetMaxAudioJitterDelay() << " ms." << endl;
     }
   }
-  PTRACE(3,"GatewayEndPoint\t Jitter buffer is "<<jitter<<" ms long.");
+  PTRACE(3, "GatewayEndPoint\t Jitter buffer is "
+         << GetMinAudioJitterDelay() << " to " << GetMaxAudioJitterDelay() << " ms");
 
 
   if (args.HasOption("tos"))
@@ -387,19 +456,37 @@ BOOL GatewayEndPoint::Initialise(PArgList & args)
   accessControl.LoadHostsAccess();
 
   H323ListenerTCP * listener;
-  if (args.GetOptionString('i').IsEmpty())
-    listener  = new H323ListenerTCP(*this);
-  else {
+
+  WORD listenPort = H323EndPoint::DefaultTcpPort;
+  if (args.HasOption("listenport"))
+    listenPort = (WORD)args.GetOptionString("listenport").AsInteger();
+
+  if (args.GetOptionString('i').IsEmpty()) {
+    PIPSocket::Address interfaceAddress(INADDR_ANY);
+    listener  = new H323ListenerTCP(*this, interfaceAddress, listenPort);
+  } else {
     PIPSocket::Address interfaceAddress(args.GetOptionString('i'));
-    listener  = new H323ListenerTCP(*this, interfaceAddress);
+    listener  = new H323ListenerTCP(*this, interfaceAddress, listenPort);
   }
 
   if (!StartListener(listener)) {
     cout <<  "Could not open H.323 listener port on "
          << listener->GetListenerPort() << endl;
+    delete listener;
     return FALSE;
   }
 
+  cout << "Listening on port " << listener->GetListenerPort() << endl;
+
+  if (args.HasOption("translate")) {
+    masqAddressPtr = new PIPSocket::Address(args.GetOptionString("translate"));
+    behind_masq = TRUE;
+    cout << "Masquerading as address " << *(masqAddressPtr) << endl;
+  } else {
+    behind_masq = FALSE;
+  }
+
+
 #ifdef HAS_IXJ
   BOOL usingIXJ = FALSE;
   if (args.HasOption('q')) {
@@ -450,10 +537,17 @@ BOOL GatewayEndPoint::Initialise(PArgList & args)
         continue;
 
       lids[i].SetLineOnHook(l);
+      PTRACE(2, "ep\tInitialise SetLineOnHook true");
       //if (!lids[i].IsLinePresent(l))
       //  cout << "No line on " << lids[i] << '-' << l << endl;
       //else {
         noLinesAvailable = FALSE;
+	if (args.HasOption("playgain")) {
+	    lids[i].SetPlayVolume(l, args.GetOptionString("playgain").AsUnsigned());
+	}
+	if (args.HasOption("recordgain")) {
+	  lids[i].SetRecordVolume(l, args.GetOptionString("recordgain").AsUnsigned());
+	}
         lids[i].SetAEC(l, aecLevel);
         if (!country)
           lids[i].SetCountryCodeName(country);
@@ -470,7 +564,8 @@ BOOL GatewayEndPoint::Initialise(PArgList & args)
         cout << endl;
 
         H323_LIDCapability::AddAllCapabilities(lids[i], capabilities, 0, 0);
-        lineMonitors.Append(new LineMonitor(*this, lids[i], l, h323Address));
+        lineMonitors.Append(new LineMonitor(*this, lids[i], l, h323Address, 
+					    args.HasOption("collect")));
       //}
     }
   }
@@ -597,7 +692,7 @@ PString GatewayEndPoint::RewriteE164(const PString &dialedDigits, const PStringT
 
 H323Connection * GatewayEndPoint::CreateConnection(unsigned callReference)
 {
-  return new GatewayConnection(*this, callReference);
+ return new GatewayConnection(*this, callReference);
 }
 
 
@@ -610,6 +705,7 @@ BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
   WORD port;
   if (address.GetIpAndPort(ip, port) && !accessControl.IsAllowed(ip)) {
     cout << "Incoming H.323 call from " << ip << " is denied." << endl;
+    PTRACE(2, "ep\tIncoming H.323 call from " << ip << " is denied.");
     return FALSE;
   }
 
@@ -626,6 +722,7 @@ BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
 
   if (myConnection.line == NULL) {
     cout << "No available lines for incoming H.323 call." << endl;
+    PTRACE(2, "ep\tOnIncomingCall::No available lines for incoming H.323 call.");
     return FALSE;
   }
 
@@ -637,6 +734,7 @@ BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
   // Sieze the line
   device.SetLineOffHook(line);
   cout << "Setting line " << device << '-' << line << " off hook" << endl;
+  PTRACE(2, "ep\tSetting line " << device << '-' << line << " off hook");
 
   PString number;
 
@@ -644,13 +742,17 @@ BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
     number = autoDial;
 
   if (!number) {
-    // Wait for dial tone (max 3 seconds)
-    if (device.WaitForTone(line, OpalLineInterfaceDevice::DialTone))
+    // Wait for dial tone
+    if (device.WaitForTone(line, OpalLineInterfaceDevice::DialTone, 5000)) {
+      PTRACE(2, "ep\t Dial tone deteted.");
       cout << "Dial tone detected." << endl;
-    else
+    } else {
+      PTRACE(2, "ep\tNO Dial tone deteted.");
       cout << "No dial tone detected." << endl;
+    }
 
     PString dialed(RewriteE164(number));
+    PTRACE(2, "ep\tDialing: " << dialed);
     cout << "Dialling: " << dialed << endl;
     device.PlayDTMF(line, dialed);
   }
@@ -658,6 +760,34 @@ BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
   return TRUE;
 }
 
+void GatewayEndPoint::TranslateTCPAddress(PIPSocket::Address &localAddr, const PIPSocket::Address &remoteAddr)
+{
+
+  if (this->behind_masq) {
+    /* Check if the remote address is a private IP address.
+     * RFC 1918 specifies the following private IP addresses
+     * 10.0.0.0    - 10.255.255.255.255
+     * 172.16.0.0  - 172.31.255.255
+     * 192.168.0.0 - 192.168.255.255
+     */
+
+     BOOL remote_address_private =
+       ( ((remoteAddr.Byte1() == 10))
+       ||((remoteAddr.Byte1() == 172)
+           && (remoteAddr.Byte2() >= 16)
+           && (remoteAddr.Byte2() <= 31) )
+       ||((remoteAddr.Byte1() == 192)
+           && (remoteAddr.Byte2() == 168)) );
+     /*
+      * We are using NAT and the remote address is outside our LAN, so
+      * replace the local address with the IP address of the NAT box.
+      */
+     if (!remote_address_private) {
+       localAddr = *(this->masqAddressPtr);
+     }
+  }
+  return;
+}
 
 H323Connection::AnswerCallResponse GatewayEndPoint::OnAnswerCall(H323Connection & connection,
                                                                  const PString & /*callerName*/,
@@ -673,15 +803,18 @@ H323Connection::AnswerCallResponse GatewayEndPoint::OnAnswerCall(H323Connection
 
   switch (device.WaitForToneDetect(line, 3000)) {
     case OpalLineInterfaceDevice::RingTone :
+      PTRACE(2, "ep\tRing tone deteted.");
       cout << "Ring tone detected." << endl;
       return H323Connection::AnswerCallNow; // Temporary code, not completed yet
 //      return H323Connection::AnswerCallPending;
 
     case OpalLineInterfaceDevice::BusyTone :
+      PTRACE(2, "ep\tBusy tone deteted, call is refused.");
       cout << "Busy tone detected, call being refused." << endl;
       break;
 
     default :
+      PTRACE(2, "ep\tNo ring/busy tone deteted.");
       cout << "No ring/busy tone detected." << endl;
   }
 
@@ -716,6 +849,7 @@ void GatewayEndPoint::OnConnectionCleared(H323Connection & connection,
   OpalLineInterfaceDevice & device = myConnection.line->GetDevice();
   unsigned lineNumber = myConnection.line->GetLineNumber();
 
+  PTRACE(2, "ep\tConnection Cleared. Set line " << device << '-' << lineNumber << " on hook.");
   cout << "Setting line " << device << '-' << lineNumber << " on hook" << endl;
   device.SetLineOnHook(lineNumber);
 }
@@ -726,7 +860,6 @@ void GatewayEndPoint::OnConnectionCleared(H323Connection & connection,
 GatewayConnection::GatewayConnection(GatewayEndPoint & _ep, unsigned callReference)
   : H323Connection(_ep, callReference)
 {
-  SetMaxAudioDelayJitter(_ep.jitter);
   line = NULL;
 }
 
@@ -746,23 +879,26 @@ void GatewayConnection::OnUserInputString(const PString & value)
 
 BOOL GatewayConnection::OnStartLogicalChannel(H323Channel & channel)
 {
-  cout << "Line " << line->GetDevice() << '-' << line->GetLineNumber()
+  PStringStream message;
+  message << "Line " << line->GetDevice() << '-' << line->GetLineNumber()
        << " started logical channel: ";
 
   switch (channel.GetDirection()) {
     case H323Channel::IsTransmitter :
-      cout << "sending ";
+      message << "sending ";
       break;
 
     case H323Channel::IsReceiver :
-      cout << "receiving ";
+      message << "receiving ";
       break;
 
     default :
       break;
   }
 
-  cout << channel.GetCapability() << endl;
+  message << channel.GetCapability();
+  cout << message << endl;
+  PTRACE(2, "Connection\t" << message);
 
   return TRUE;
 }
@@ -773,12 +909,14 @@ BOOL GatewayConnection::OnStartLogicalChannel(H323Channel & channel)
 LineMonitor::LineMonitor(GatewayEndPoint & ep,
                          OpalLineInterfaceDevice & dev,
                          unsigned line,
-                         const PString & addr)
+                         const PString & addr,
+			 BOOL collectNum)
   : PThread(10000, NoAutoDeleteThread),
     endpoint(ep),
     device(dev),
     lineNumber(line),
-    h323Address(addr)
+    h323Address(addr),
+    collectNumber(collectNum)
 {
   Resume();
 }
@@ -795,6 +933,9 @@ static BOOL IsLIDRinging(OpalLineInterfaceDevice & lid, PINDEX line, PTimer & ti
   if (timer.IsRunning())
     return TRUE;
 
+  if (lid.IsLineOffHook(line))    // The Phone cannot be ringing, cause it is off hook.
+    return FALSE;
+
   if (!lid.IsLineRinging(line))
     return FALSE;
 
@@ -813,8 +954,9 @@ void LineMonitor::Main()
   while (!shutdown.Wait(500)) {
 
     // if the line is ringing, start a new call
+    PTRACE(3, "PSTNGW\tLine Monitor, start of loop");
     if (IsLIDRinging(device, lineNumber, ringTimer)) {
-
+      PTRACE(3, "PSTNGW\tLine Monitor, detected line is ringing");
       cout << "Line " << device << '-' << lineNumber << " is ringing, ";
 
       // ignore incoming calls when no IP address to dial
@@ -827,18 +969,27 @@ void LineMonitor::Main()
         while (!shutdown.Wait(500) && IsLIDRinging(device, lineNumber, ringTimer))
           ;
         cout << "Ringing stopped on device " << device << "-" << lineNumber << endl;
+	PTRACE(3, "PSTNGW\tRinging stopped on device " << device << "-" << lineNumber);
       }
       else {
+	PTRACE(3, "PSTNGW\tLine Monitor. Call address " << h323Address);
         cout << "calling \"" << h323Address << "\"." << endl;
 
         // if still onhook, wait for the ringing to stop
-        if (HandleCall()) 
-          cout << "Call from " << device << '-' << lineNumber << " to " << h323Address << " ended." << endl;
-        else {
-          cout << "Call from " << device << '-' << lineNumber << " to " << h323Address << " failed." << endl;
+	PStringStream message;
+        if (HandleCall()) {
+	  message << "Call from " << device << '-' << lineNumber << " to " << h323Address << " ended.";
+	  PTRACE(3, "PSTNGW\tLine Monitor:" << message);
+          cout << message << endl;
+        } else {
+	  message << "Call from " << device <<  '-' << lineNumber << " to " << h323Address << " failed.";
+	  PTRACE(3, "PSTNGW\tLine Monitor: " << message);
+	  cout << message << endl;
+
           while (!shutdown.Wait(500) && IsLIDRinging(device, lineNumber, ringTimer))
             ;
           cout << "Ringing stopped on device " << device << "-" << lineNumber << endl;
+	  PTRACE(3, "PSTNGW\tLine Monitor, Ringing stopped on device" << device << '-' << lineNumber);
         }
       }
     }
@@ -846,6 +997,7 @@ void LineMonitor::Main()
     // if a call is active, but the line has been disconnected, hangup the call
     if (!callToken && device.IsLineDisconnected(lineNumber)) {
       cout << "Line " << device << '-' << lineNumber << " disconnected." << endl;
+      PTRACE(3, "PSTNGW\tLine Monitor, Line " << device << '-' << lineNumber << " disconnected");
       endpoint.ClearCall(callToken);
       callToken = PString();
     }
@@ -857,46 +1009,134 @@ void LineMonitor::Main()
 
 BOOL LineMonitor::HandleCall()
 {
-  PString callerID;
+  PString   callerID;
+  PString   exth323Address;
+  const int max_digits = 10;
+  char      number[max_digits];
+  
+  *number = 0;
+
   if (device.GetCallerID(lineNumber, callerID))
     cout << "Caller is " << callerID << endl;
 
-  GatewayConnection * myConnection = (GatewayConnection *)endpoint.MakeCall(h323Address, callToken);
-  if (myConnection == NULL)
-    return FALSE;
+  if (collectNumber) {
+    // DR - collect number to dial before we make H323 call
+    // - take lid port off-hook
+    // - generate dial tone from lid
+    // - collect digits until # pressed or 10 second timeout
+    // - every new digit resets timer for another 10 seconds
 
-  myConnection->line = this;
-  myConnection->SetLocalPartyName(callerID);
-  // The myConnection variable becomes suspect after this, so don't use it
+    device.SetLineOffHook(lineNumber);
+    device.PlayTone(lineNumber, OpalLineInterfaceDevice::DialTone);
+
+    char      digit;
+    int       num_digits = 0;
+
+    PTimer collectTimer(10000);
+
+    cout << "Enter number to dial at other end, then press #" << endl;
+    do {
+      digit = device.ReadDTMF(lineNumber);
+      if (digit) {
+	number[num_digits++] = digit;
+	cout << "digit: " << digit << endl;
+	collectTimer = 10000;
+      }
+      if (device.IsLineDisconnected(lineNumber)) {
+        cerr << "Caller hung up. " << endl;
+        return FALSE;
+      }
+      if (shutdown.Wait(50))
+        return FALSE;
+    } while((digit != '#') && (num_digits < max_digits) && 
+	    collectTimer.IsRunning());
+
+    device.StopTone(lineNumber);
+
+    if (num_digits) {
+      if (digit == '#')
+	num_digits--;
+      number[num_digits] = 0;
+      cout << "num_digits: " << num_digits << " number: " << number << endl;
+    }
+    else {
+      device.SetLineOnHook(lineNumber);
+      return FALSE;
+    }
+  }
+
+  // if we have collected a number concatenate with ip host
+  if (*number) {
+    exth323Address = PString(number) + PString("@") + h323Address;
+  }
+  else {
+    exth323Address = h323Address;    
+  }
+  cout << "Make call to h323 Address " << exth323Address << endl;
+  PTRACE(2, "Monitor\tMake call to h323 Address " << exth323Address);
+  // now make call
+
+  {
+    GatewayConnection * myConnection = (GatewayConnection *)endpoint.MakeCallLocked(exth323Address, callToken);
+    if (myConnection == NULL)
+      return FALSE;
+
+    myConnection->line = this;
+    myConnection->SetLocalPartyName(callerID);
+    myConnection->Unlock();
+  }
 
   PTimer ringTimer(3000);
 
   while (!endpoint.IsConnectionEstablished(callToken)) {
     // If H.323 call has failed then ignore the call
-    if (!endpoint.HasConnection(callToken))
+    if (!endpoint.HasConnection(callToken)) {
+      cerr << "endpoint does not have connection for " << callToken << endl;
       return FALSE;
+    }
 
-    // If the ringing stops before establishment, clear the call and exit
-    if (!IsLIDRinging(device, lineNumber, ringTimer)) {
+    if (!collectNumber) {
+      // If the ringing stops before establishment, clear the call and exit
+      if (!IsLIDRinging(device, lineNumber, ringTimer)) {
+	cerr << "Lid is not ringing. " << endl;
+	endpoint.ClearCall(callToken);
+	return FALSE;
+      } 
+    }
+
+    if (device.IsLineDisconnected(lineNumber)) {
+      cerr << "Caller hung up. " << endl;
       endpoint.ClearCall(callToken);
       return FALSE;
     }
+
+    if (shutdown.Wait(50))
+      return FALSE;
   }
 
-  cout << "Line " << device << '-' << lineNumber << " going off hook" << endl;
-
-  // Have H.323 call answered, so answer the PSTN line
-  device.SetLineOffHook(lineNumber);
-
+  if (!collectNumber) {
+    cout << "Line " << device << '-' << lineNumber << " going off hook" <<endl;
+    PTRACE(2, "Monitor\tLine " << device << '-' << lineNumber << " going off hook");
+    // Have H.323 call answered, so answer the PSTN line
+    device.SetLineOffHook(lineNumber);
+  } 
+  
   // Now wait for either party to hang up
   while (endpoint.HasConnection(callToken) && !device.IsLineDisconnected(lineNumber)) 
     if (shutdown.Wait(200))
       return TRUE;
 
   cout << "Line " << device << '-' << lineNumber << " disconnected." << endl;
+  PTRACE(2, "Monitor\tLine " << device << '-' << lineNumber << " disconnected.");
   endpoint.ClearCall(callToken);
   return TRUE;
 }
 
 
 // End of File ///////////////////////////////////////////////////////////////
+
+
+
+
+
+
